from typing import List, Dict
import csv
import nasdaqdatalink

# Define the function to load indicators directly in this file to avoid circular imports
def load_indicator_metadata(filepath="src/resources/world_data_bank/metadata/metadata.csv") -> Dict[str, str]:
    """Load indicator metadata from CSV file."""
    mapping = {}
    with open(filepath, newline="") as f:
        reader = csv.reader(f)
        for row in reader:
            if len(row) >= 2:
                series_id, description = row[0].strip(), row[1].strip()
                if series_id and description:
                    mapping[series_id] = description
    return mapping

# Load indicators on module import
INDICATORS = load_indicator_metadata()


def get_indicator_value(country: str, indicator: str) -> str:
    """
    Fetch the most recent value of a World Bank development indicator for a given country.

    The `indicator` should describe what you're interested in. Examples:
    - "share of youth not in education"
    - "employment to population ratio"
    - "female NEET rate"
    - "GDP per person employed"
    """
    try:
        df = nasdaqdatalink.get_table(
            "WB/DATA",
            series_id=indicator,
            country_code=country,
        )
        if df.empty:
            return f"No data found for {indicator} in country."
        return df
    except Exception as e:
        return f"Error fetching {indicator}: {str(e)}"


def search_indicators(keyword: str) -> List[str]:
    """
    Search for indicator descriptions matching a given keyword.
    """
    keyword_lower = keyword.lower()
    matches = [
        f"{code}: {desc}"
        for code, desc in INDICATORS.items()
        if keyword_lower in desc.lower()
    ]
    return matches[:10]  # return up to 10 matches


def list_all_indicators() -> List[str]:
    """
    List all available indicator codes and descriptions.
    """
    return [f"{series_id}: {desc}" for series_id, desc in INDICATORS.items()]



from typing import Union, Optional, List, Dict
import nasdaqdatalink
import pandas as pd


def get_company_stats(
    symbol: Optional[str] = None,
    figi: Optional[str] = None
) -> Union[pd.DataFrame, str]:
    """
    Fetch company statistics from Nasdaq Data Link E360 NDAQ/STAT table.
    
    Args:
        symbol: Stock ticker symbol (e.g., 'AAPL', 'MSFT')
        figi: Bloomberg FIGI identifier (e.g., 'BBG000BPH459')
        
    Returns:
        DataFrame containing company statistics or error message
    """
    if not symbol and not figi:
        return "Error: Either symbol or figi must be provided."
    
    try:
        params = {}
        if symbol:
            params["symbol"] = symbol
        if figi:
            params["figi"] = figi
            
        # Fetch data from NDAQ/STAT table
        data = nasdaqdatalink.get_table('NDAQ/STAT', **params)
        
        if data.empty:
            return f"No data found for the specified criteria."
        
        return data
    except Exception as e:
        return f"Error fetching company statistics: {str(e)}"


def list_available_fields() -> List[Dict[str, str]]:
    """
    List available fields in the NDAQ/STAT table with descriptions.
    
    Returns:
        List of dictionaries containing field name and description
    """
    fields = [
        {"name": "symbol", "description": "Symbol of the company"},
        {"name": "figi", "description": "Unique Identifier given by Bloomberg"},
        {"name": "marketcap", "description": "Market cap of the security calculated as shares outstanding * previous day close"},
        {"name": "high52week", "description": "Highest fully adjusted price observed during trading hours over the past 52 calendar weeks"},
        {"name": "high52week_date", "description": "The date when the highest fully adjusted price was observed over the past 52 weeks"},
        {"name": "low52week", "description": "Lowest fully adjusted price observed during trading hours over the past 52 calendar weeks"},
        {"name": "low52week_date", "description": "The date when the lowest fully adjusted price was observed over the past 52 weeks"},
        {"name": "avgvolume1m", "description": "Average 1 month volume based on calendar days"},
        {"name": "avgvolume3m", "description": "Average 3-month volume based on calendar days"},
        {"name": "divyield", "description": "Dividend Yield measures the ratio between a company's dividendPerShare and its price"},
        {"name": "dividendpershare", "description": "The dividends per share unit, adjusted for stock splits and stock dividends"},
        {"name": "nextdividenddate", "description": "Expected ex date of the next dividend"},
        {"name": "exdividenddate", "description": "Ex date of the last dividend"},
        {"name": "nextearningsdate", "description": "Expected next earnings report date"},
        {"name": "pe", "description": "Measures the ratio between marketcap and netinccmn"},
        {"name": "epsdil", "description": "Earnings per diluted share as calculated and reported by the company"},
        {"name": "eps", "description": "Earnings per share as calculated and reported by the company"},
        {"name": "pe1", "description": "An alternative to pe representing the ratio between price and eps"},
        {"name": "pb", "description": "Measures the ratio between marketcap and equity"},
        {"name": "freefloat", "description": "The free float percentage of shares available for trading (0-1)"}
    ]
    return fields


from typing import Union, Optional, List, Dict, Any
import nasdaqdatalink
import pandas as pd


def get_corporate_actions(
    symbol: Optional[str] = None,
    figi: Optional[str] = None,
    date: Optional[str] = None,
    action: Optional[str] = None
) -> Union[pd.DataFrame, str]:
    """
    Fetch corporate actions data from Nasdaq Data Link E360 NDAQ/CA table.
    
    Args:
        symbol: Stock ticker symbol (e.g., 'AAPL', 'MSFT')
        figi: Bloomberg FIGI identifier (e.g., 'BBG000BPH459')
        date: Date of the corporate action in YYYY-MM-DD format
        action: Type of corporate action (e.g., 'split', 'merger')
        
    Returns:
        DataFrame containing corporate actions data or error message
    """
    try:
        params = {}
        if symbol:
            params["symbol"] = symbol
        if figi:
            params["figi"] = figi
        if date:
            params["date"] = date
        if action:
            params["action"] = action
            
        # Fetch data from NDAQ/CA table
        data = nasdaqdatalink.get_table('NDAQ/CA', **params)
        
        if data.empty:
            return f"No corporate action data found for the specified criteria."
        
        return data
    except Exception as e:
        return f"Error fetching corporate actions data: {str(e)}"


def list_available_corporate_action_fields() -> List[Dict[str, Any]]:
    """
    List available fields in the NDAQ/CA table with descriptions.
    
    Returns:
        List of dictionaries containing field name, description, and type
    """
    fields = [
        {"name": "date", "description": "The date of the corporate action", "type": "Date", "filterable": True, "primary_key": True},
        {"name": "symbol", "description": "Symbol of the company", "type": "String", "filterable": True, "primary_key": True},
        {"name": "figi", "description": "Unique Identifier given by Bloomberg", "type": "String", "filterable": True},
        {"name": "action", "description": "Corporate action type (e.g., 'split', 'merger')", "type": "String", "primary_key": True},
        {"name": "value", "description": "The total USD value of the action", "type": "Double"},
        {"name": "contrasymbol", "description": "The contra symbol of the action, or N/A if not applicable", "type": "String", "primary_key": True},
        {"name": "contraname", "description": "The name of the contra company, or N/A if not applicable", "type": "String", "primary_key": True}
    ]
    return fields


from typing import List, Dict, Optional
from mcp.server.fastmcp import FastMCP
from nasdaq_data_link_mcp_os.resources.common.countries import get_country_code
from nasdaq_data_link_mcp_os.resources.world_data_bank.indicators import (
    get_indicator_value as get_wb_indicator,
    search_indicators,
    list_all_indicators,
)
from nasdaq_data_link_mcp_os.resources.rtat.retail_activity import (
    get_rtat10_data,
    get_rtat_data,
)
from nasdaq_data_link_mcp_os.resources.equities_360.statistics import (
    get_company_stats,
    list_available_fields,
)
from nasdaq_data_link_mcp_os.resources.equities_360.fundamentals import (
    get_fundamental_summary,
    list_available_fundamental_fields,
)
from nasdaq_data_link_mcp_os.resources.equities_360.fundamental_details import (
    get_fundamental_details,
    list_available_detail_fields,
)
from nasdaq_data_link_mcp_os.resources.equities_360.balance_sheet import (
    get_balance_sheet,
    list_available_balance_sheet_fields,
)
from nasdaq_data_link_mcp_os.resources.equities_360.cash_flow import (
    get_cash_flow,
    list_available_cash_flow_fields,
)
from nasdaq_data_link_mcp_os.resources.equities_360.corporate_actions import (
    get_corporate_actions,
    list_available_corporate_action_fields,
)
from nasdaq_data_link_mcp_os.resources.equities_360.reference_data import (
    get_reference_data,
    list_available_reference_fields,
)
from config import initialize_api

# Initialize API configuration
initialize_api()

mcp = FastMCP("NASDAQ Data Link MCP", dependencies=["nasdaq-data-link", "pycountry"])


@mcp.tool()
def get_indicator_value(country: str, indicator: str) -> str:
    return get_wb_indicator(country, indicator)


@mcp.tool()
def country_code(countryName: str) -> str:
    return get_country_code(countryName)


@mcp.tool()
def list_worldbank_indicators() -> List[str]:
    return list_all_indicators()


@mcp.tool()
def search_worldbank_indicators(keyword: str) -> List[str]:
    return search_indicators(keyword)


@mcp.tool()
def get_rtat10(dates: str, tickers: str = None):
    """
    Retrieves Retail Trading Activity Tracker 10 (RTAT10) data for specific dates and optional tickers.

    Example: get_rtat10(dates='2025-03-31,2025-03-28,2025-03-27', tickers='TSLA,TQQQ,SQQQ')
    """
    return get_rtat10_data(dates, tickers)


@mcp.tool()
def get_rtat(dates: str, tickers: str = None):
    """
    Retrieves Retail Trading Activity (RTAT) data for specific dates and optional tickers.

    Example: get_rtat(dates='2025-03-31,2025-03-28,2025-03-27', tickers='TSLA,TQQQ,SQQQ')
    """
    return get_rtat_data(dates, tickers)


@mcp.tool()
def get_stock_stats(symbol: Optional[str] = None, figi: Optional[str] = None):
    """
    Retrieves company statistics from Nasdaq Equities 360 database.
    
    Provides comprehensive statistics for a company including market cap, PE ratio, 
    52-week highs/lows, dividend information, and more.
    
    Either symbol or figi must be provided.
    
    Example: get_stock_stats(symbol='MSFT')
    Example: get_stock_stats(figi='BBG000BPH459')
    """
    return get_company_stats(symbol, figi)


@mcp.tool()
def list_stock_stat_fields() -> List[Dict[str, str]]:
    """
    Lists all available fields in the stock statistics database with descriptions.
    
    This helps users understand what data is available through the get_stock_stats tool.
    """
    return list_available_fields()


@mcp.tool()
def get_fundamental_data(symbol: Optional[str] = None, figi: Optional[str] = None, calendardate: Optional[str] = None, dimension: Optional[str] = None):
    """
    Retrieves fundamental financial data from Nasdaq Equities 360 Fundamental Summary database.
    
    Provides comprehensive fundamental data including profitability ratios, valuation metrics,
    income statement items, and financial health indicators.
    
    Either symbol or figi must be provided.
    
    Parameters:
      - symbol: Stock ticker symbol (e.g., 'MSFT')
      - figi: Bloomberg FIGI identifier (e.g., 'BBG000BPH459')
      - calendardate: Calendar date in YYYY-MM-DD format (e.g., '2022-12-31')
      - dimension: Data dimension (MRQ: quarterly, MRY: annual, MRT: trailing twelve months)
    
    Example: get_fundamental_data(symbol='MSFT', dimension='MRY')
    Example: get_fundamental_data(figi='BBG000BPH459', calendardate='2022-12-31')
    """
    return get_fundamental_summary(symbol, figi, calendardate, dimension)


@mcp.tool()
def list_fundamental_fields() -> List[Dict[str, str]]:
    """
    Lists all available fields in the fundamental summary database with descriptions.
    
    This helps users understand what data is available through the get_fundamental_data tool,
    including profitability ratios, valuation metrics, and financial health indicators.
    """
    return list_available_fundamental_fields()


@mcp.tool()
def get_detailed_financials(symbol: Optional[str] = None, figi: Optional[str] = None, calendardate: Optional[str] = None, dimension: Optional[str] = None):
    """
    Retrieves detailed financial data from Nasdaq Equities 360 Fundamental Details database.
    
    Provides comprehensive financial statement data including balance sheet items, income statement 
    components, cash flow statement details, and financial ratios.
    
    Either symbol or figi must be provided.
    
    Parameters:
      - symbol: Stock ticker symbol (e.g., 'MSFT')
      - figi: Bloomberg FIGI identifier (e.g., 'BBG000BPH459')
      - calendardate: Calendar date in YYYY-MM-DD format (e.g., '2022-12-31')
      - dimension: Data dimension (MRQ: quarterly, MRY: annual, MRT: trailing twelve months)
    
    Example: get_detailed_financials(symbol='MSFT', dimension='MRY')
    Example: get_detailed_financials(figi='BBG000BPH459', calendardate='2022-12-31')
    """
    return get_fundamental_details(symbol, figi, calendardate, dimension)


@mcp.tool()
def list_detailed_financial_fields() -> List[Dict[str, str]]:
    """
    Lists all available fields in the fundamental details database with descriptions.
    
    This helps users understand what data is available through the get_detailed_financials tool,
    including balance sheet items, income statement components, cash flow details, and more.
    """
    return list_available_detail_fields()


@mcp.tool()
def get_balance_sheet_data(symbol: Optional[str] = None, figi: Optional[str] = None, calendardate: Optional[str] = None, dimension: Optional[str] = None):
    """
    Retrieves balance sheet data from Nasdaq Equities 360 Balance Sheet database.
    
    Provides comprehensive balance sheet data including assets, liabilities, equity, 
    and detailed breakdowns of each category.
    
    Either symbol or figi must be provided.
    
    Parameters:
      - symbol: Stock ticker symbol (e.g., 'MSFT')
      - figi: Bloomberg FIGI identifier (e.g., 'BBG000BPH459')
      - calendardate: Calendar date in YYYY-MM-DD format (e.g., '2022-12-31')
      - dimension: Data dimension (MRQ: quarterly, MRY: annual, MRT: trailing twelve months)
    
    Example: get_balance_sheet_data(symbol='MSFT', dimension='MRY')
    Example: get_balance_sheet_data(figi='BBG000BPH459', calendardate='2022-12-31')
    """
    return get_balance_sheet(symbol, figi, calendardate, dimension)


@mcp.tool()
def list_balance_sheet_fields() -> List[Dict[str, str]]:
    """
    Lists all available fields in the balance sheet database with descriptions.
    
    This helps users understand what data is available through the get_balance_sheet_data tool,
    including assets, liabilities, equity, and detailed breakdowns of each category.
    """
    return list_available_balance_sheet_fields()


@mcp.tool()
def get_cash_flow_data(symbol: Optional[str] = None, figi: Optional[str] = None, calendardate: Optional[str] = None, dimension: Optional[str] = None):
    """
    Retrieves cash flow statement data from Nasdaq Equities 360 Cash Flow database.
    
    Provides cash flow statement data including operating, investing, and financing
    activities, as well as free cash flow and capital expenditure information.
    
    Either symbol or figi must be provided.
    
    Parameters:
      - symbol: Stock ticker symbol (e.g., 'MSFT')
      - figi: Bloomberg FIGI identifier (e.g., 'BBG000BPH459')
      - calendardate: Calendar date in YYYY-MM-DD format (e.g., '2022-12-31')
      - dimension: Data dimension (MRQ: quarterly, MRY: annual, MRT: trailing twelve months)
    
    Example: get_cash_flow_data(symbol='MSFT', dimension='MRY')
    Example: get_cash_flow_data(figi='BBG000BPH459', calendardate='2022-12-31')
    """
    return get_cash_flow(symbol, figi, calendardate, dimension)


@mcp.tool()
def list_cash_flow_fields() -> List[Dict[str, str]]:
    """
    Lists all available fields in the cash flow statement database with descriptions.
    
    This helps users understand what data is available through the get_cash_flow_data tool,
    including operating, investing, and financing cash flows, and related metrics.
    """
    return list_available_cash_flow_fields()


@mcp.tool()
def get_corporate_action_data(symbol: Optional[str] = None, figi: Optional[str] = None, date: Optional[str] = None, action: Optional[str] = None):
    """
    Retrieves corporate actions data from Nasdaq Equities 360 Corporate Actions database.
    
    Provides information about corporate events such as stock splits, mergers, acquisitions,
    and other significant company actions that can affect stock price and ownership.
    
    Parameters:
      - symbol: Stock ticker symbol (e.g., 'TSLA')
      - figi: Bloomberg FIGI identifier
      - date: Date of the corporate action in YYYY-MM-DD format (e.g., '2023-03-24')
      - action: Type of corporate action (e.g., 'split', 'merger')
    
    Example: get_corporate_action_data(symbol='TSLA', action='split')
    Example: get_corporate_action_data(date='2023-03-24')
    """
    return get_corporate_actions(symbol, figi, date, action)


@mcp.tool()
def list_corporate_action_fields() -> List[Dict[str, str]]:
    """
    Lists all available fields in the corporate actions database with descriptions.
    
    This helps users understand what data is available through the get_corporate_action_data tool,
    including date, action type, value, and related company information.
    """
    return list_available_corporate_action_fields()


@mcp.tool()
def get_company_reference_data(symbol: Optional[str] = None, figi: Optional[str] = None):
    """
    Retrieves company reference data from Nasdaq Equities 360 Reference Data database.
    
    Provides static information about companies including exchange, industry, sector,
    company website, SEC filings links, and location information.
    
    Either symbol or figi must be provided.
    
    Parameters:
      - symbol: Stock ticker symbol (e.g., 'AMD')
      - figi: Bloomberg FIGI identifier (e.g., 'BBG000BBQCY0')
    
    Example: get_company_reference_data(symbol='AMD')
    Example: get_company_reference_data(figi='BBG000BBQCY0')
    """
    return get_reference_data(symbol, figi)


@mcp.tool()
def list_reference_data_fields() -> List[Dict[str, str]]:
    """
    Lists all available fields in the company reference database with descriptions.
    
    This helps users understand what data is available through the get_company_reference_data tool,
    including exchange, industry, sector, company website, and location information.
    """
    return list_available_reference_fields()



from typing import List, Optional
import nasdaqdatalink


def get_rtat10_data(dates: str, tickers: Optional[str] = None):
    """
    Fetch Retail Trading Activity Tracker 10 (RTAT10) data for specific dates and tickers.

    Args:
        dates: Comma-separated list of dates in format 'YYYY-MM-DD'
        tickers: Optional comma-separated list of ticker symbols

    Returns:
        DataFrame with RTAT10 data or error message
    """
    try:
        params = {"date": dates}
        if tickers:
            params["ticker"] = tickers

        df = nasdaqdatalink.get_table("NDAQ/RTAT10", **params)

        if df.empty:
            return "No RTAT10 data found for the specified parameters."
        return df
    except Exception as e:
        return f"Error fetching RTAT10 data: {str(e)}"


def get_rtat_data(dates: str, tickers: Optional[str] = None):
    """
    Fetch Retail Trading Activity (RTAT) data for specific dates and tickers.

    Args:
        dates: Comma-separated list of dates in format 'YYYY-MM-DD'
        tickers: Optional comma-separated list of ticker symbols

    Returns:
        DataFrame with RTAT data or error message
    """
    try:
        params = {"date": dates}
        if tickers:
            params["ticker"] = tickers

        df = nasdaqdatalink.get_table("NDAQ/RTAT", **params)

        if df.empty:
            return "No RTAT data found for the specified parameters."
        return df
    except Exception as e:
        return f"Error fetching RTAT data: {str(e)}"




from typing import Union, Optional, List, Dict, Any
import nasdaqdatalink
import pandas as pd


def get_balance_sheet(
    symbol: Optional[str] = None,
    figi: Optional[str] = None,
    calendardate: Optional[str] = None,
    dimension: Optional[str] = None
) -> Union[pd.DataFrame, str]:
    """
    Fetch balance sheet data from Nasdaq Data Link E360 NDAQ/BS table.
    
    Args:
        symbol: Stock ticker symbol (e.g., 'AAPL', 'MSFT')
        figi: Bloomberg FIGI identifier (e.g., 'BBG000BPH459')
        calendardate: Calendar date in YYYY-MM-DD format
        dimension: MRQ (Quarterly), MRY (Annual), or MRT (Trailing-twelve-months)
        
    Returns:
        DataFrame containing balance sheet data or error message
    """
    if not symbol and not figi:
        return "Error: Either symbol or figi must be provided."
    
    try:
        params = {}
        if symbol:
            params["symbol"] = symbol
        if figi:
            params["figi"] = figi
        if calendardate:
            params["calendardate"] = calendardate
        if dimension:
            params["dimension"] = dimension
            
        # Fetch data from NDAQ/BS table
        data = nasdaqdatalink.get_table('NDAQ/BS', **params)
        
        if data.empty:
            return f"No data found for the specified criteria."
        
        return data
    except Exception as e:
        return f"Error fetching balance sheet data: {str(e)}"


def list_available_balance_sheet_fields() -> List[Dict[str, Any]]:
    """
    List available fields in the NDAQ/BS table with descriptions.
    
    Returns:
        List of dictionaries containing field name, description, and type
    """
    fields = [
        {"name": "calendardate", "description": "The Calendar Date represents the normalized reportperiod", "type": "Date", "filterable": True, "primary_key": True},
        {"name": "symbol", "description": "Symbol of the company", "type": "String", "filterable": True, "primary_key": True},
        {"name": "figi", "description": "Unique Identifier given by Bloomberg", "type": "String", "filterable": True},
        {"name": "reportperiod", "description": "The Report Period represents the end date of the fiscal period", "type": "Date", "primary_key": True},
        {"name": "dimension", "description": "Dimensional view of data (MRQ: Quarterly, MRY: annual, MRT: trailing-twelve-months)", "type": "String", "filterable": True, "primary_key": True},
        {"name": "assets", "description": "Total assets recognized on the balance sheet", "type": "BigInt"},
        {"name": "bvps", "description": "Book Value Per Share (ratio between equity and shareswa)", "type": "Double"},
        {"name": "cashneq", "description": "Cash and equivalents on hand", "type": "BigInt"},
        {"name": "debt", "description": "Total current and non-current debt", "type": "BigInt"},
        {"name": "liabilities", "description": "Total liabilities recognized on the balance sheet", "type": "BigInt"},
        {"name": "equity", "description": "Total stockholders' equity", "type": "BigInt"},
        {"name": "accoci", "description": "Accumulated other comprehensive income", "type": "BigInt"},
        {"name": "currency", "description": "The company functional reporting currency", "type": "String"},
        {"name": "investmentsnc", "description": "Non-current portion of investments", "type": "BigInt"},
        {"name": "investmentsc", "description": "Current portion of investments", "type": "BigInt"},
        {"name": "receivables", "description": "Trade and non-trade receivables", "type": "BigInt"},
        {"name": "inventory", "description": "Inventory expected to be sold within one year", "type": "BigInt"},
        {"name": "assetsnc", "description": "Non-current assets", "type": "BigInt"},
        {"name": "assetsc", "description": "Current assets", "type": "BigInt"},
        {"name": "investments", "description": "Total marketable and non-marketable securities and other invested assets", "type": "BigInt"},
        {"name": "ppnenet", "description": "Property, plant, and equipment (net of depreciation)", "type": "BigInt"},
        {"name": "intangibles", "description": "Intangible assets and goodwill", "type": "BigInt"},
        {"name": "tangibles", "description": "Tangible assets (assets minus intangibles)", "type": "BigInt"},
        {"name": "payables", "description": "Trade and non-trade payables", "type": "BigInt"},
        {"name": "liabilitiesnc", "description": "Non-current liabilities", "type": "BigInt"},
        {"name": "liabilitiesc", "description": "Current liabilities", "type": "BigInt"},
        {"name": "debtc", "description": "Current portion of debt", "type": "BigInt"},
        {"name": "debtnc", "description": "Non-current portion of debt", "type": "BigInt"},
        {"name": "deferredrev", "description": "Deferred revenue (unrecognized revenue from received payments)", "type": "BigInt"},
        {"name": "retearn", "description": "Retained earnings (cumulative undistributed earnings/deficit)", "type": "BigInt"},
        {"name": "taxassets", "description": "Tax assets and receivables", "type": "BigInt"},
        {"name": "taxliabilities", "description": "Outstanding tax liabilities", "type": "BigInt"},
        {"name": "cashnequsd", "description": "Cash and equivalents in USD", "type": "BigInt"}
    ]
    return fields


from typing import Union, Optional, List, Dict, Any
import nasdaqdatalink
import pandas as pd


def get_fundamental_details(
    symbol: Optional[str] = None,
    figi: Optional[str] = None,
    calendardate: Optional[str] = None,
    dimension: Optional[str] = None
) -> Union[pd.DataFrame, str]:
    """
    Fetch detailed fundamental data from Nasdaq Data Link E360 NDAQ/FD table.
    
    Args:
        symbol: Stock ticker symbol (e.g., 'AAPL', 'MSFT')
        figi: Bloomberg FIGI identifier (e.g., 'BBG000BPH459')
        calendardate: Calendar date in YYYY-MM-DD format
        dimension: MRQ (Quarterly), MRY (Annual), or MRT (Trailing-twelve-months)
        
    Returns:
        DataFrame containing detailed fundamental data or error message
    """
    if not symbol and not figi:
        return "Error: Either symbol or figi must be provided."
    
    try:
        params = {}
        if symbol:
            params["symbol"] = symbol
        if figi:
            params["figi"] = figi
        if calendardate:
            params["calendardate"] = calendardate
        if dimension:
            params["dimension"] = dimension
            
        # Fetch data from NDAQ/FD table
        data = nasdaqdatalink.get_table('NDAQ/FD', **params)
        
        if data.empty:
            return f"No data found for the specified criteria."
        
        return data
    except Exception as e:
        return f"Error fetching fundamental details: {str(e)}"


def list_available_detail_fields() -> List[Dict[str, Any]]:
    """
    List available fields in the NDAQ/FD table with descriptions.
    
    Returns:
        List of dictionaries containing field name, description, and type
    """
    fields = [
        {"name": "calendardate", "description": "The Calendar Date represents the normalized reportperiod", "type": "Date", "filterable": True, "primary_key": True},
        {"name": "symbol", "description": "Symbol of the company", "type": "String", "filterable": True, "primary_key": True},
        {"name": "figi", "description": "Unique Identifier given by Bloomberg", "type": "String", "filterable": True},
        {"name": "reportperiod", "description": "The Report Period represents the end date of the fiscal period", "type": "Date", "primary_key": True},
        {"name": "dimension", "description": "Dimensional view of data (MRQ: Quarterly, MRY: annual, MRT: trailing-twelve-months)", "type": "String", "filterable": True, "primary_key": True},
        {"name": "payables", "description": "A component of liabilities representing trade and non-trade payables", "type": "BigInt"},
        {"name": "receivables", "description": "A component of assets representing trade and non-trade receivables", "type": "BigInt"},
        {"name": "cashneq", "description": "A component of assets representing currency on hand and demand deposits", "type": "BigInt"},
        {"name": "investmentsnc", "description": "The non-current portion of investments", "type": "BigInt"},
        {"name": "investmentsc", "description": "The current portion of investments", "type": "BigInt"},
        {"name": "taxassets", "description": "A component of assets representing tax assets and receivables", "type": "BigInt"},
        {"name": "ncfo", "description": "Cash inflow (outflow) from operating activities", "type": "BigInt"},
        {"name": "capex", "description": "Cash inflow (outflow) associated with acquisition & disposal of long-lived assets", "type": "BigInt"},
        {"name": "cashnequsd", "description": "Cash and equivalents in USD", "type": "BigInt"},
        {"name": "netincdis", "description": "Income from a disposal group, net of income tax", "type": "BigInt"},
        {"name": "debtc", "description": "The current portion of debt", "type": "BigInt"},
        {"name": "dps", "description": "Dividends declared per split-adjusted share of common stock", "type": "Double"},
        {"name": "retearn", "description": "Cumulative amount of undistributed earnings or deficit", "type": "BigInt"},
        {"name": "ebitda", "description": "Earnings Before Interest, Taxes, Depreciation and Amortization", "type": "BigInt"},
        {"name": "ebit", "description": "Earnings Before Interest and Tax", "type": "BigInt"},
        {"name": "equity", "description": "Total stockholders' equity", "type": "BigInt"},
        {"name": "ncfbus", "description": "Cash inflow (outflow) from acquisition & disposal of businesses", "type": "BigInt"},
        {"name": "depamor", "description": "Depreciation, amortization, and accretion", "type": "BigInt"},
        {"name": "taxexp", "description": "Current and deferred income tax expense for continuing operations", "type": "BigInt"},
        {"name": "sgna", "description": "Selling, general and administrative expenses", "type": "BigInt"},
        {"name": "rnd", "description": "Research and development expenses", "type": "BigInt"},
        {"name": "opex", "description": "Operating expenses (sgna, rnd and other operating expenses)", "type": "BigInt"},
        {"name": "sbcomp", "description": "Non-cash, equity-based employee remuneration", "type": "BigInt"},
        {"name": "tangibles", "description": "Value of tangible assets (assets minus intangibles)", "type": "BigInt"},
        {"name": "intangibles", "description": "Carrying amounts of intangible assets and goodwill", "type": "BigInt"},
        {"name": "eps", "description": "Earnings per share as reported by the company", "type": "Double"},
        {"name": "ebt", "description": "Earnings Before Tax", "type": "BigInt"},
        {"name": "netinc", "description": "Net income attributable to the parent", "type": "BigInt"},
        {"name": "opinc", "description": "Operating income (before interest, taxes and non-operating items)", "type": "BigInt"},
        {"name": "consolinc", "description": "Consolidated net income", "type": "BigInt"},
        {"name": "inventory", "description": "Value of inventory expected to be sold within one year", "type": "BigInt"},
        {"name": "debtnc", "description": "The non-current portion of debt", "type": "BigInt"},
        {"name": "taxliabilities", "description": "Outstanding tax liabilities", "type": "BigInt"},
        {"name": "liabilitiesnc", "description": "The non-current portion of liabilities", "type": "BigInt"},
        {"name": "liabilities", "description": "Total liabilities", "type": "BigInt"},
        {"name": "deferredrev", "description": "Consideration received but not recognized as revenue", "type": "BigInt"},
        {"name": "accoci", "description": "Accumulated other comprehensive income", "type": "BigInt"},
        {"name": "ppnenet", "description": "Property, plant, and equipment (net of depreciation)", "type": "BigInt"},
        {"name": "pe1", "description": "Price to earnings ratio (price/eps)", "type": "Double"},
        {"name": "gp", "description": "Gross profit (revenue minus cost of revenue)", "type": "BigInt"},
        {"name": "grossmargin", "description": "Ratio between gross profit and revenue", "type": "Double"},
        {"name": "revenue", "description": "Total revenue from goods sold and services rendered", "type": "BigInt"},
        {"name": "intexp", "description": "Interest expense on borrowed funds", "type": "BigInt"},
        {"name": "cor", "description": "Cost of revenue (goods produced and sold, services rendered)", "type": "BigInt"},
        {"name": "ncfcommon", "description": "Cash flow from common equity changes", "type": "BigInt"},
        {"name": "netinccmn", "description": "Net income attributable to common shareholders", "type": "BigInt"},
        {"name": "ncfinv", "description": "Cash flow from investments", "type": "BigInt"},
        {"name": "ncff", "description": "Cash flow from financing activities", "type": "BigInt"},
        {"name": "shareswa", "description": "Weighted average shares outstanding used to calculate EPS", "type": "BigInt"},
        {"name": "epsusd", "description": "Earnings per share in USD", "type": "Double"},
        {"name": "workingcapital", "description": "Difference between current assets and current liabilities", "type": "BigInt"},
        {"name": "invcap", "description": "Invested capital", "type": "BigInt"},
        {"name": "invcapavg", "description": "Average invested capital for the period", "type": "BigInt"},
        {"name": "roic", "description": "Return on Invested Capital (EBIT/invcapavg)", "type": "Double"},
        {"name": "freecashflow", "description": "Free Cash Flow (ncfo minus capex)", "type": "BigInt"},
        {"name": "ebitdamargin", "description": "Ratio between EBITDA and revenue", "type": "Double"},
        {"name": "ev", "description": "Enterprise value (marketcap plus debt minus cashneq)", "type": "BigInt"},
        {"name": "evebit", "description": "Enterprise Value to EBIT ratio", "type": "BigInt"},
        {"name": "evebitda", "description": "Enterprise Value to EBITDA ratio", "type": "Double"},
        {"name": "divyield", "description": "Dividend Yield (dps/price)", "type": "Double"},
        {"name": "assetturnover", "description": "Revenue divided by average assets", "type": "Double"},
        {"name": "revenueusd", "description": "Revenue in USD", "type": "BigInt"},
        {"name": "debttoassets", "description": "Cash flow from debt issuance/repayment", "type": "BigInt"},
        {"name": "shareswadil", "description": "Weighted average diluted shares outstanding", "type": "BigInt"},
        {"name": "sharesbas", "description": "Basic shares outstanding after stock splits", "type": "BigInt"}
    ]
    return fields


from typing import Union, Optional, List, Dict, Any
import nasdaqdatalink
import pandas as pd


def get_cash_flow(
    symbol: Optional[str] = None,
    figi: Optional[str] = None,
    calendardate: Optional[str] = None,
    dimension: Optional[str] = None
) -> Union[pd.DataFrame, str]:
    """
    Fetch cash flow statement data from Nasdaq Data Link E360 NDAQ/CF table.
    
    Args:
        symbol: Stock ticker symbol (e.g., 'AAPL', 'MSFT')
        figi: Bloomberg FIGI identifier (e.g., 'BBG000BPH459')
        calendardate: Calendar date in YYYY-MM-DD format
        dimension: MRQ (Quarterly), MRY (Annual), or MRT (Trailing-twelve-months)
        
    Returns:
        DataFrame containing cash flow statement data or error message
    """
    if not symbol and not figi:
        return "Error: Either symbol or figi must be provided."
    
    try:
        params = {}
        if symbol:
            params["symbol"] = symbol
        if figi:
            params["figi"] = figi
        if calendardate:
            params["calendardate"] = calendardate
        if dimension:
            params["dimension"] = dimension
            
        # Fetch data from NDAQ/CF table
        data = nasdaqdatalink.get_table('NDAQ/CF', **params)
        
        if data.empty:
            return f"No data found for the specified criteria."
        
        return data
    except Exception as e:
        return f"Error fetching cash flow statement data: {str(e)}"


def list_available_cash_flow_fields() -> List[Dict[str, Any]]:
    """
    List available fields in the NDAQ/CF table with descriptions.
    
    Returns:
        List of dictionaries containing field name, description, and type
    """
    fields = [
        {"name": "calendardate", "description": "The Calendar Date represents the normalized reportperiod", "type": "Date", "filterable": True, "primary_key": True},
        {"name": "symbol", "description": "Symbol of the company", "type": "String", "filterable": True, "primary_key": True},
        {"name": "figi", "description": "Unique Identifier given by Bloomberg", "type": "String", "filterable": True},
        {"name": "reportperiod", "description": "The Report Period represents the end date of the fiscal period", "type": "Date", "primary_key": True},
        {"name": "dimension", "description": "Dimensional view of data (MRQ: Quarterly, MRY: annual, MRT: trailing-twelve-months)", "type": "String", "filterable": True, "primary_key": True},
        {"name": "opex", "description": "Operating expenses (excluding cost of revenue)", "type": "BigInt"},
        {"name": "ncfi", "description": "Net cash flow from investing activities", "type": "BigInt"},
        {"name": "ncff", "description": "Net cash flow from financing activities", "type": "BigInt"},
        {"name": "fcf", "description": "Free Cash Flow (ncfo minus capex)", "type": "BigInt"},
        {"name": "ncfo", "description": "Net cash flow from operating activities", "type": "BigInt"},
        {"name": "capex", "description": "Capital expenditure (acquisition of long-lived assets)", "type": "BigInt"},
        {"name": "ncfbus", "description": "Net cash flow from business acquisitions/disposals", "type": "BigInt"},
        {"name": "sbcomp", "description": "Stock-based compensation", "type": "BigInt"},
        {"name": "depamor", "description": "Depreciation, amortization, and accretion", "type": "BigInt"},
        {"name": "ncfcommon", "description": "Net cash flow from common equity changes", "type": "BigInt"},
        {"name": "ncfinv", "description": "Net cash flow from investments", "type": "BigInt"},
        {"name": "debttoassets", "description": "Net cash flow from debt issuance/repayment", "type": "BigInt"}
    ]
    return fields


from typing import Union, Optional, List, Dict, Any
import nasdaqdatalink
import pandas as pd


def get_fundamental_summary(
    symbol: Optional[str] = None,
    figi: Optional[str] = None,
    calendardate: Optional[str] = None,
    dimension: Optional[str] = None
) -> Union[pd.DataFrame, str]:
    """
    Fetch fundamental summary data from Nasdaq Data Link E360 NDAQ/FS table.
    
    Args:
        symbol: Stock ticker symbol (e.g., 'AAPL', 'MSFT')
        figi: Bloomberg FIGI identifier (e.g., 'BBG000BPH459')
        calendardate: Calendar date in YYYY-MM-DD format
        dimension: MRQ (Quarterly), MRY (Annual), or MRT (Trailing-twelve-months)
        
    Returns:
        DataFrame containing fundamental data or error message
    """
    if not symbol and not figi:
        return "Error: Either symbol or figi must be provided."
    
    try:
        params = {}
        if symbol:
            params["symbol"] = symbol
        if figi:
            params["figi"] = figi
        if calendardate:
            params["calendardate"] = calendardate
        if dimension:
            params["dimension"] = dimension
            
        # Fetch data from NDAQ/FS table
        data = nasdaqdatalink.get_table('NDAQ/FS', **params)
        
        if data.empty:
            return f"No data found for the specified criteria."
        
        return data
    except Exception as e:
        return f"Error fetching fundamental summary: {str(e)}"


def list_available_fundamental_fields() -> List[Dict[str, Any]]:
    """
    List available fields in the NDAQ/FS table with descriptions.
    
    Returns:
        List of dictionaries containing field name, description, and type
    """
    fields = [
        {"name": "calendardate", "description": "The Calendar Date represents the normalized reportperiod", "type": "Date", "filterable": True, "primary_key": True},
        {"name": "symbol", "description": "Symbol of the company", "type": "String", "filterable": True, "primary_key": True},
        {"name": "figi", "description": "Unique Identifier given by Bloomberg", "type": "String", "filterable": True},
        {"name": "reportperiod", "description": "The Report Period represents the end date of the fiscal period", "type": "Date", "primary_key": True},
        {"name": "dimension", "description": "Dimensional view of data (MRQ: Quarterly, MRY: annual, MRT: trailing-twelve-months)", "type": "String", "filterable": True, "primary_key": True},
        {"name": "fcfps", "description": "Free Cash Flow per Share", "type": "Double"},
        {"name": "ps", "description": "Price to Sales ratio between marketcap and revenue", "type": "Double"},
        {"name": "pe", "description": "Price to Earnings ratio between marketcap and netinccmn", "type": "Double"},
        {"name": "revenue", "description": "Amount of Revenue recognized from goods sold, services rendered, etc.", "type": "BigInt"},
        {"name": "currentratio", "description": "The ratio between assetsc and liabilitiesc", "type": "Double"},
        {"name": "de", "description": "Debt to Equity ratio between liabilities and equity", "type": "Double"},
        {"name": "roa", "description": "Return on Assets (netinccmn relative to total assets)", "type": "Double"},
        {"name": "roe", "description": "Return on Equity (netinccmn as percentage of equityavg)", "type": "Double"},
        {"name": "ros", "description": "Return on Sales (ebit divided by revenue)", "type": "Double"},
        {"name": "gp", "description": "Gross Profit (revenue less cost of revenue)", "type": "BigInt"},
        {"name": "opinc", "description": "Operating income (before deduction of intexp, taxexp, etc.)", "type": "BigInt"},
        {"name": "netmargin", "description": "Net Margin (ratio between netinccmn and revenue)", "type": "Double"},
        {"name": "ebitda", "description": "Earnings Before Interest, Taxes, Depreciation and Amortization", "type": "BigInt"},
        {"name": "bvps", "description": "Book Value Per Share (ratio between equity and shareswa)", "type": "Double"},
        {"name": "evebit", "description": "Enterprise Value to EBIT ratio", "type": "BigInt"},
        {"name": "evebitda", "description": "Enterprise Value to EBITDA ratio", "type": "Double"},
        {"name": "tbvps", "description": "Tangible Book Value Per Share", "type": "Double"}
    ]
    return fields


from typing import Union, Optional, List, Dict, Any
import nasdaqdatalink
import pandas as pd


def get_reference_data(
    symbol: Optional[str] = None,
    figi: Optional[str] = None
) -> Union[pd.DataFrame, str]:
    """
    Fetch reference data from Nasdaq Data Link E360 NDAQ/RD table.
    
    Args:
        symbol: Stock ticker symbol (e.g., 'AAPL', 'MSFT')
        figi: Bloomberg FIGI identifier (e.g., 'BBG000BPH459')
        
    Returns:
        DataFrame containing reference data or error message
    """
    try:
        params = {}
        if symbol:
            params["symbol"] = symbol
        if figi:
            params["figi"] = figi
            
        # Fetch data from NDAQ/RD table
        data = nasdaqdatalink.get_table('NDAQ/RD', **params)
        
        if data.empty:
            return f"No reference data found for the specified criteria."
        
        return data
    except Exception as e:
        return f"Error fetching reference data: {str(e)}"


def list_available_reference_fields() -> List[Dict[str, Any]]:
    """
    List available fields in the NDAQ/RD table with descriptions.
    
    Returns:
        List of dictionaries containing field name, description, and type
    """
    fields = [
        {"name": "symbol", "description": "Symbol of the company", "type": "String", "filterable": True, "primary_key": True},
        {"name": "figi", "description": "Unique Identifier given by Bloomberg", "type": "String", "filterable": True},
        {"name": "exchange", "description": "The exchange on which the security trades (e.g., 'NASDAQ', 'NYSE')", "type": "String"},
        {"name": "category", "description": "Security category (e.g., 'Domestic Common Stock')", "type": "String"},
        {"name": "cusips", "description": "Security identifier (space delimited for multiple identifiers)", "type": "String"},
        {"name": "name", "description": "Full legal name of the company", "type": "String"},
        {"name": "industry", "description": "Industry classification based on SIC codes", "type": "String"},
        {"name": "companysite", "description": "URL of the company website", "type": "String"},
        {"name": "sector", "description": "Sector classification based on SIC codes", "type": "String"},
        {"name": "sector_1", "description": "Primary sector exposure (TIIC level 1)", "type": "String"},
        {"name": "sector_2", "description": "Secondary sector exposure (TIIC level 1)", "type": "String"},
        {"name": "secfilings", "description": "URL to SEC filings with Central Index Key (CIK)", "type": "String"},
        {"name": "siccode", "description": "Standard Industrial Classification code", "type": "Integer"},
        {"name": "location", "description": "Company location as registered with the SEC", "type": "String"}
    ]
    return fields